# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from pyramid.httpexceptions import HTTPUnauthorized


@pytest.mark.usefixtures('rollback_registry')
class TestCredential:

    def init_user(self, registry):
        self.User = registry.Pyramid.User
        self.CredentialStore = registry.Pyramid.CredentialStore
        self.User.insert(login='user.1')
        self.CredentialStore.insert(login='user.1', password="P1")

    def test_check_login_ok(self, rollback_registry):
        self.init_user(rollback_registry)
        assert self.User.check_login(login="user.1", password="P1") == 'user.1'

    def test_check_login_ko(self, rollback_registry):
        self.init_user(rollback_registry)
        with pytest.raises(HTTPUnauthorized):
            self.User.check_login(login="user.1", password="P2")



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2002 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.view import view_config, view_defaults
from anyblok_pyramid import current_blok


@view_defaults(installed_blok=current_blok())
class Bloks:
    def __init__(self, request):
        self.request = request
        self.registry = request.anyblok.registry

    @view_config(route_name='bloks', renderer="json",
                 request_method='GET', permission="read")
    def get_all(self):
        query = self.registry.Pyramid.restrict_query_by_user(
            self.registry.System.Blok.query(),
            self.request.authenticated_userid
        )
        bloks = query.all()
        return bloks.to_dict('name', 'author', 'version')

    @view_config(route_name='blok', renderer="json",
                 request_method='GET', permission="read")
    def get_one(self):
        blok = self.registry.System.Blok.query().filter_by(
            name=self.request.matchdict['name']).one()
        return blok.to_dict('name', 'author', 'version')

    @view_config(route_name='blok', renderer="json",
                 request_method='PUT', permission="write")
    def put_one(self):
        blok = self.registry.System.Blok.query().filter_by(
            name=self.request.matchdict['name']).one()
        return blok.to_dict('name', 'author', 'version')



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok


def declarations(reload=None):
    from . import user
    from . import role
    if reload:
        reload(user)
        reload(role)


class Auth(Blok):
    version = "0.1.0"
    author = "Jean-Sébastien Suzanne"
    required = ['anyblok-core', 'pyramid']

    @classmethod
    def import_declaration_module(cls):
        declarations()

    @classmethod
    def reload_declaration_module(cls, reload):
        declarations(reload=reload)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import json
import pytest
from anyblok.config import Configuration
from unittest import mock
from urllib.parse import urlparse, parse_qs


try:
    from anyblok_pyramid.bloks.pyramid import oidc
    has_oidc = True
except ImportError:
    has_oidc = False


class MockJsonResponse:

    headers = {"content-type": "application/json"}

    def __init__(self, data, status_code):
        self.data = data
        self.status_code = status_code

    @property
    def text(self):
        return json.dumps(self.data)


def mock_request(method=None, url=None, *args, **kwargs):
    if url == "http://fake/.well-known/openid-configuration":
        return MockJsonResponse(
            {
                "issuer": "http://fake",
                "authorization_endpoint": "http://fake/oauth/authorize",
                "token_endpoint": "http://fake/oauth/token",
                "revocation_endpoint": "https://fake/oauth/revoke",
                "introspection_endpoint": "https://fake/oauth/introspect",
                "userinfo_endpoint": "https://fake/oauth/userinfo",
                "jwks_uri": "https://fake/oauth/discovery/keys",
                "scopes_supported": ["openid", "profile", "email"],
                "response_types_supported": ["code", "token"],
                "response_modes_supported": ["query", "fragment"],
                "token_endpoint_auth_methods_supported": [
                    "client_secret_basic",
                    "client_secret_post",
                ],
                "subject_types_supported": ["public"],
                "id_token_signing_alg_values_supported": ["RS256"],
                "claim_types_supported": ["normal"],
                "claims_supported": ["name", "email", "profile"],
            },
            200,
        )
    elif url == "http://fake/oauth/token":
        qs = parse_qs(kwargs["data"])
        assert qs["redirect_uri"][0] == "http://localhost/oidc_callback"
        return MockJsonResponse(
            {
                "access_token": "a_test_token"
                if qs["code"][0] == "a-fake-code"
                else "other_token",
                "token_type": "token",
                "scope": ["openid email"],
                "state": qs["state"][0],
                "expires_in": 0,
                "refresh_token": "",
            },
            200,
        )
    elif url == "https://fake/oauth/userinfo":
        qs = parse_qs(kwargs["data"])
        if qs["access_token"][0] == "a_test_token":
            return MockJsonResponse(
                {"custom_userinfo_field": "user@anyblok.org"}, 200
            )
        else:
            return MockJsonResponse(
                {"custom_userinfo_field": "unkwnon_user@anyblok.org"}, 200
            )
    else:
        raise Exception("Unexpected url: {}".format(url))


class TestPyramidBlokBase:
    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok, webserver):
        transaction = registry_testblok.begin_nested()

        def try_rollback(*args, **kwargs):
            """Wrap rollback to be silent in case where transaction is
            expectedly already rollback"""

            try:
                transaction.rollback(*args, **kwargs)
            except Exception:
                # if transaction is already rollback do not raise exceptions
                pass

        request.addfinalizer(try_rollback)

    @pytest.fixture(autouse=True)
    def reset_webserver(self, request, registry_testblok, webserver):
        def clear_auth():
            webserver.reset()

        request.addfinalizer(clear_auth)
        return


class TestPyramidBlok(TestPyramidBlokBase):
    def test_current_blok(self, registry_testblok, webserver):
        registry = registry_testblok
        webserver.get("/hello/JS/", status=404)
        registry.upgrade(install=("test-pyramid1",))
        resp = webserver.get("/hello/JS/", status=200)
        assert resp.body.decode("utf8") == "Hello JS !!!"

    def test_simple_crud_security(self, registry_testblok, webserver):
        registry = registry_testblok
        registry.upgrade(install=("test-pyramid2",))
        webserver.get("/bloks", status=403)
        webserver.get("/blok/auth", status=403)
        resp = webserver.post_json(
            "/login", {"login": "viewer", "password": ""}, status=302
        )
        headers = resp.headers
        webserver.get("/bloks", status=200, headers=headers)
        webserver.get("/blok/auth", status=200, headers=headers)
        webserver.put("/blok/auth", {}, status=403, headers=headers)
        resp = webserver.post("/logout", {}, status=302)
        headers = resp.headers
        webserver.get("/bloks", status=403, headers=headers)
        webserver.get("/blok/auth", status=403, headers=headers)
        resp = webserver.post_json(
            "/login", {"login": "admin", "password": ""}, status=302
        )
        headers = resp.headers
        webserver.get("/bloks", status=200, headers=headers)
        webserver.get("/blok/auth", status=200, headers=headers)
        webserver.put("/blok/auth", {}, status=200, headers=headers)

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    def test_no_issuer(self, registry_testblok, webserver):
        self.assert_missing_config(registry_testblok, "oidc_provider_issuer")

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    def assert_missing_config(self, registry, config):
        registry.upgrade(install=("test-pyramid2",))
        with pytest.raises(ValueError) as ex:
            oidc.get_client()
        assert config in str(
            ex.value
        ), "if no oidc provider issuer is set an exception must raises"

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    def test_no_client_id(self, registry_testblok, webserver):
        Configuration.set("oidc_provider_issuer", "http://fake")
        self.assert_missing_config(
            registry_testblok, "oidc_relying_party_client_id"
        )

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    def test_no_secret_id(self, registry_testblok, webserver):
        Configuration.set("oidc_provider_issuer", "http://fake")
        Configuration.set("oidc_relying_party_client_id", "abc")
        self.assert_missing_config(
            registry_testblok, "oidc_relying_party_secret_id"
        )

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    def test_no_rp_callback(self, registry_testblok, webserver):
        Configuration.set("oidc_provider_issuer", "http://fake")
        Configuration.set("oidc_relying_party_client_id", "abc")
        Configuration.set("oidc_relying_party_secret_id", "efg")
        self.assert_missing_config(
            registry_testblok, "oidc_relying_party_callback"
        )

    def oidc_common(self, registry, webserver):
        Configuration.set("oidc_provider_issuer", "http://fake")
        Configuration.set("oidc_relying_party_client_id", "test_client_id")
        Configuration.set("oidc_relying_party_secret_id", "test_secret_id")
        Configuration.set(
            "oidc_relying_party_callback", "http://localhost/oidc_callback"
        )
        SCOPE = "test1,test2"
        Configuration.set("oidc_scope", SCOPE)
        Configuration.set("oidc_userinfo_field", "custom_userinfo_field")
        registry.upgrade(install=("test-pyramid2",))
        resp = webserver.get("/bloks", status=403)
        webserver.get("/blok/auth", status=403)
        resp = webserver.get("/oidc_login", status=302)
        url = urlparse(resp.headers.get("Location"))
        qs = parse_qs(url.query, strict_parsing=True)
        assert url.hostname == "fake"
        assert url.path == "/oauth/authorize"
        assert qs["client_id"][0] == "test_client_id"
        assert qs["response_type"][0] == "code"
        assert qs["scope"][0] == SCOPE.replace(",", " ")
        assert qs["redirect_uri"][0] == "http://localhost/oidc_callback"
        assert qs["state"][0]
        assert qs["nonce"][0]
        return qs

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    @mock.patch("requests.request", side_effect=mock_request)
    def test_oidc_auth(self, mock_oidc, registry_testblok, webserver):
        qs = self.oidc_common(registry_testblok, webserver)
        # user is redirect to OIDC provider in order to do the authentication
        # he comes back to the oidc_callback uri with a code and the current
        # state
        curerent_cookie = webserver.cookies["None"]
        webserver.get(
            "/oidc_callback?code=a-fake-code&state={}".format(qs["state"][0]),
            status=302,
        )
        assert webserver.cookies["None"] != curerent_cookie
        webserver.get("/bloks", status=200)

    @pytest.mark.skipif(not has_oidc, reason="oic is not installed")
    @mock.patch("requests.request", side_effect=mock_request)
    def test_unkown_user_oidc_auth(
        self, mock_oidc, registry_testblok, webserver
    ):
        qs = self.oidc_common(registry_testblok, webserver)
        curerent_cookie = webserver.cookies["None"]
        webserver.get(
            "/oidc_callback?code=another-fake-code&state={}".format(
                qs["state"][0]
            ),
            status=401,
        )
        assert webserver.cookies["None"] == curerent_cookie
        webserver.get("/bloks", status=403)


class TestPyramidBlokRestrictQueryByUserId(TestPyramidBlokBase):
    def test_restrict_query_by_user_not_restricted(
        self, registry_testblok, webserver
    ):
        registry = registry_testblok
        registry.upgrade(install=("test-pyramid2",))
        resp = webserver.post_json(
            "/login", {"login": "viewer", "password": ""}, status=302
        )
        headers = resp.headers
        results = webserver.get("/bloks", status=200, headers=headers)
        assert len(results.json) > 1

    def test_restrict_query_by_user_restricted(
        self, registry_testblok, webserver
    ):
        registry = registry_testblok
        registry.upgrade(install=("test-pyramid2",))
        resp = webserver.post_json(
            "/login", {"login": "user2@anyblok.org", "password": ""}, status=302
        )
        headers = resp.headers
        results = webserver.get("/bloks", status=200, headers=headers)
        assert len(results.json) == 1



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import BlokManager
import inspect
from .common import (  # noqa: F401
    merge,
    PERM_READ,
    PERM_WRITE,
    PERM_C___,
    PERM__R__,
    PERM___U_,
    PERM____D,
)


def anyblok_init_config(unittest=False):
    from anyblok import config  # noqa import anyblok.config
    from . import config  # noqa import config definition


class AnyBlokPyramidException(Exception):
    pass


def current_blok():
    filename = inspect.stack()[1][1]
    for blok in BlokManager.ordered_bloks:
        if filename.startswith(BlokManager.getPath(blok)):
            return blok

    raise AnyBlokPyramidException("You are not in a Blok")  # pragma: no cover



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2019 Alexis TOURNEUX <tourneuxalexis@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String
from anyblok.field import Function


@Declarations.register(Declarations.Model.Pyramid)
class User:
    """User declaration for identity"""

    first_name = String(nullable=False)
    last_name = String(nullable=False)
    name = Function(fget='get_name')

    def get_name(self):
        """Return the name of the user"""
        return self.first_name + ' ' + self.last_name.upper()



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2002 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.common import add_autodocs
from anyblok.model.plugins import ModelPluginBase


class RestrictQueryByUserIdPlugin(ModelPluginBase):
    """An AnyBlok plugin that helps to add extra filters on queries according
    the current user"""

    def __init__(self, registry):
        if not hasattr(registry, "restrict_query_by_user_methods"):
            registry.restrict_query_by_user_methods = {}

        super().__init__(registry)

    def transform_base_attribute(
        self, attr, method, namespace, base, transformation_properties,
        new_type_properties,
    ):
        """Find restricted methods in the base to save the
        namespace and the method in the registry
        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        if (
            hasattr(method, "is_a_restric_query_by_user_method")
            and method.is_a_restric_query_by_user_method is True
        ):
            if namespace not in self.registry.restrict_query_by_user_methods:
                self.registry.restrict_query_by_user_methods[namespace] = {}
            self.registry.restrict_query_by_user_methods[namespace].update(
                {method.__name__: method}
            )


def restrict_query_by_user():
    autodoc = """Decorator to register a class method to restrict query
    by user. Decorated class method will receive a query object and a user
    instance, developer must return a query object with additionnal filters::

        from anyblok_pyramid.bloks.pyramid.restrict import (
            restrict_query_by_user
        )
        ...
            @restrict_query_by_user()
            def restrict_reading_this_blok_to_user2(cls, query, user):
                if user.login == "user2@anyblok.org":
                    query = query.filter_by(name="test-pyramid2")
                return query
    """

    def wrapper(method):
        add_autodocs(method, autodoc)
        method.is_a_restric_query_by_user_method = True
        return classmethod(method)

    return wrapper



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Alexis TOURNEUX <tourneuxalexis@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from datetime import datetime, date, timedelta
from pyramid.renderers import JSON
from uuid import UUID, uuid1
from os import urandom
from decimal import Decimal
from anyblok_pyramid.adapter import (
    datetime_adapter,
    date_adapter,
    TimeDeltaModes,
    timedelta_adapter_factory,
    uuid_adapter,
    bytes_adapter,
    decimal_adapter,
)
from anyblok_pyramid.testing import init_web_server
import pytest


class TestAdapter:

    def test_registry_get_datetime(self):

        def get_datetime(request):
            return {'datetime': datetime(2017, 10, 1, 1, 1, 1)}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_datetime, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(datetime, datetime_adapter)
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)
        assert res.json_body['datetime'] == datetime_adapter(
            datetime(2017, 10, 1, 1, 1, 1), None)

    def test_registry_get_timedelta(self):

        def get_timedelta(request):
            return {'timedelta': timedelta(
                days=1, hours=4, minutes=56,
                seconds=3710, milliseconds=4000,
                microseconds=500)}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(
                get_timedelta, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(
                timedelta, timedelta_adapter_factory())
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)

        assert res.json_body['timedelta'] == timedelta(
            days=1, hours=4, minutes=56, seconds=3710,
            milliseconds=4000, microseconds=500).total_seconds()

        # This part is aimed at testing that the different modes implemented on
        # timedelta adapter are behaving as intented
        delta = timedelta(
            weeks=1, days=3, hours=15, minutes=3, seconds=45, milliseconds=354,
            microseconds=768
        )

        assert round(
            timedelta_adapter_factory(TimeDeltaModes.SECONDS)(
                delta, None), 8) == 918225.354768
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.MICROSECONDS)(
                delta, None), 8) == 918225354768
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.MILLISECONDS)(
                delta, None), 8) == 918225354.768
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.MINUTES)(
                delta, None), 8) == 15303.7559128
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.HOURS)(
                delta, None), 8) == 255.06259855
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.DAYS)(
                delta, None), 8) == 10.62760827
        assert round(
            timedelta_adapter_factory(TimeDeltaModes.WEEKS)(
                delta, None), 8) == 1.51822975

        with pytest.raises(ValueError):
            timedelta_adapter_factory("not a valid mode")(delta, None)

    def test_registry_get_date(self):

        def get_date(request):
            return {'date': date(2017, 10, 1)}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_date, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(date, date_adapter)
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)
        assert res.json_body['date'] == date_adapter(date(2017, 10, 1), None)

    def test_registry_get_uuid(self):

        uuid = uuid1()

        def get_uuid(request):
            return {'uuid': uuid}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_uuid, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(UUID, uuid_adapter)
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)
        assert res.json_body['uuid'] == uuid_adapter(uuid, None)

    def test_registry_get_bytes(self):

        val = urandom(100)

        def get_bytes(request):
            return {'bytes': val}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_bytes, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(bytes, bytes_adapter)
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)
        assert res.json_body['bytes'] == bytes_adapter(val, None)

    def test_registry_get_decimal(self):

        val = Decimal('100.12')

        def get_bytes(request):
            return {'decimal': val}

        def add_route_and_views(config):
            config.add_route('dbname', '/test/')
            config.add_view(get_bytes, route_name='dbname', renderer='json')
            json_renderer = JSON()
            json_renderer.add_adapter(Decimal, decimal_adapter)
            config.add_renderer('json', json_renderer)

        webserver = init_web_server(add_route_and_views)
        res = webserver.get('/test/', status=200)
        assert res.json_body['decimal'] == decimal_adapter(val, None)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from ..exceptions import RecursionRoleError


@pytest.mark.usefixtures('rollback_registry')
class TestUserAndRole:

    def init_user(self, registry):
        self.User = registry.Pyramid.User
        self.Role = registry.Pyramid.Role
        self.user = self.User.insert(
            login='user.1')
        self.role = self.Role.insert(name='admin', label="Administrator")
        self.role.users.append(self.user)

    def test_role_name_in_user_roles(self, rollback_registry):
        self.init_user(rollback_registry)
        assert self.role.name in self.User.get_roles(self.user.login)

    def test_rec_roles(self, rollback_registry):
        self.init_user(rollback_registry)
        role2 = self.Role.insert(name='g2', label="G2")
        self.role.children.append(role2)
        role2.children.append(self.role)
        with pytest.raises(RecursionRoleError):
            rollback_registry.flush()

    def test_check_user_exists_ok(self, rollback_registry):
        Pyramid = rollback_registry.Pyramid
        self.init_user(rollback_registry)
        assert self.user is Pyramid.check_user_exists('user.1')

    def test_check_user_exists_ko(self, rollback_registry):
        Pyramid = rollback_registry.Pyramid
        with pytest.raises(KeyError) as err:
            Pyramid.check_user_exists('user.1')
        assert str(err.value) == "'user.1 is not a valid login'"



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from copy import deepcopy
from anyblok_pyramid import merge
from anyblok.testing import LogCapture, tmp_configuration
from anyblok.config import Configuration
from anyblok_pyramid.common import get_registry_for, preload_databases
from sqlalchemy.exc import ResourceClosedError
from logging import INFO, WARNING


class TestCommon:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()

        def rollback():
            try:
                transaction.rollback()
            except ResourceClosedError:
                pass

        request.addfinalizer(rollback)
        return

    def test_get_registry_for(self, registry_testblok):
        registry = get_registry_for(Configuration.get('db_name'))
        assert registry is not None

    def test_preload_databases(self):
        db_name = Configuration.get('db_name')
        with tmp_configuration(db_names=[db_name]):
            with LogCapture('anyblok_pyramid.common', level=INFO) as handler:
                preload_databases()
                messages = handler.get_info_messages()
                assert messages
                assert 'The database %r is preloaded' % db_name in messages

    def test_preload_unexisting_databases(self):
        db_name = 'wrong_db_name'
        with tmp_configuration(db_names=[db_name]):
            with LogCapture('anyblok_pyramid.common', level=WARNING) as handler:
                preload_databases()
                messages = handler.get_warning_messages()
                assert messages
                assert 'The database %r does not exist' % db_name in messages


@pytest.mark.parametrize(
    "d1,d2,expected",
    [
        (
            {"a": {"b": {"c": "d"}}},
            {"a": {"b": {"e": "f"}}},
            {"a": {"b": {"c": "d", "e": "f"}}},
        ),
        (
            {"a": {"b": {"c": "d"}}},
            {},
            {"a": {"b": {"c": "d"}}},
        ),
        (
            {},
            {"a": {"b": {"c": "d"}}},
            {"a": {"b": {"c": "d"}}},
        ),
        (
            {"a": 1},
            {"a": 2},
            {"a": 2},
        ),
        (
            {"a": 1},
            {"a": {"b": "c"}},
            {"a": {"b": "c"}},
        ),
        (
            {"a": {"b": "c"}},
            {"a": 1},
            {"a": 1},
        ),
        (
            {"a": {"b": {"c": "d"}}},
            {"a": {"c": 1}},
            {"a": {"b": {"c": "d"}, "c": 1}},
        ),
        (
            {"a": {"c": 1}},
            {"a": {"b": {"c": "d"}}},
            {"a": {"b": {"c": "d"}, "c": 1}},
        ),
        (
            {"a": {"b": {"c": "d", "e": "z", "g": "h"}}},
            {"a": {"b": {"e": "f"}}},
            {"a": {"b": {"c": "d", "e": "f", "g": "h"}}},
        ),
        (
            {"a": {"b": ["c", "d"]}},
            {"a": {"b": ["e", "f"]}},
            {"a": {"b": ["c", "d", "e", "f"]}},
        ),
        (
            {"a": {"b": ["c"]}},
            {"a": 1},
            {"a": 1},
        ),
        (
            {"a": 1},
            {"a": {"b": ["c"]}},
            {"a": {"b": ["c"]}},
        ),
        (
            {"a": {"b": {"c": "d"}}},
            {"a": {"b": ["c"]}},
            {"a": {"b": ["c"]}},
        ),
        (
            {"a": {"b": ["c"]}},
            {"a": {"b": {"c": "d"}}},
            {"a": {"b": {"c": "d"}}},
        ),
        (
            {"a": {"b": ["c"]}},
            {"a": {"c": 1}},
            {"a": {"b": ["c"], "c": 1}},
        ),
        (
            {"a": {"c": 1}},
            {"a": {"b": ["c"]}},
            {"a": {"b": ["c"], "c": 1}},
        ),
    ],
)
def test_merge(d1, d2, expected):
    copy_d1 = deepcopy(d1)
    copy_d2 = deepcopy(d2)
    assert merge(d1, d2) == expected
    assert d1 == copy_d1
    assert d2 == copy_d2



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized
from anyblok import Declarations
try:
    from pyramid.authorization import Allow, ALL_PERMISSIONS
except ImportError:
    from pyramid.security import Allow, ALL_PERMISSIONS


@Declarations.register(Declarations.Model)
class Pyramid:

    @classmethod
    def get_roles(cls, login):
        """Return the roles of an user

        This method must be ober writting by the auth blok

        :param login: str, login attribute of the user
        :rtype: list of str (name of the roles)
        """
        return []

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Retun the ACL for a ressource and a user

        Auth, does not implement any rule to compute ACL,
        This method allow all user to use the resource ask
        by controllers.

        For other configuration, this method must be overwrite

        This method must be ober writting by the auth blok

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param params: all options need to compute ACL
        """
        # cache the method
        return [(Allow, login, ALL_PERMISSIONS)]

    @classmethod
    def check_acl(cls, login, resource, type_):
        """Retun True if user is allowed to make action type
        of the resource

        This method must be ober writting by the auth blok

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param type: str, name of the action
        :param params: all options need to compute ACL
        """
        # cache the method
        return True

    @classmethod
    def format_login_params(cls, request):
        """Return the login and password from query

        By default the query come from json_body and are named
        **login** and **password**

        If the entries come from another place, this method must be overwrite
        :param request: the request from the controllers
        """
        return request.json_body

    @classmethod
    def check_login(cls, **kwargs):
        """Check login / password

        This method raise an exception, because any credential
        is stored in this bloks

        .. warning::

            This method must be overwriting by anycredential blok


        :param kwargs: any options need to validate credential
        """
        raise HTTPUnauthorized()

    @classmethod
    def check_user_exists(cls, login):
        raise NotImplementedError()

    @classmethod
    def get_user(cls, user_id):
        """Return user for a given `user_id`, to be overwrite in user bloks
        The method is called by `get_user` cached method to retreive user
        while restricting query by user.

        .. warning::

            User cache invalidation must be done by developer.

        :param user_id: This is the user primary key (could be a loging
            according the User class definition)
        :return User (instance ?) or None: retreive the user for the given
            `user_id`
        """
        return user_id

    @classmethod
    def restrict_query_by_user(
        cls, query, user_code
    ):
        """Call registered decorated method (by
        ``from anyblok_pyramid.bloks.pyramid.restrict.restrict_query_by_user``)
        to add filters on current query according the selected model.

        :param query: A query object which you have to add filters
        :user_code: User primary key value used to retreive users.

        .. note::

            This method is using get_user which cached user instance, you
            have to manage or mind to cache invalidation while using this
            method.
        """
        for method in cls.anyblok.restrict_query_by_user_methods.get(
            query.Model, []
        ):

            query = getattr(
                cls.anyblok.get(query.Model), method
            )(query, cls.get_user(user_code))

        return query



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from wsgiref.simple_server import make_server
from anyblok.blok import BlokManager
from anyblok.config import Configuration
from .pyramid_config import Configurator
import hupper
import sys
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)
from .common import preload_databases
from logging import getLogger

logger = getLogger(__name__)


def wsgi():
    """Simple Pyramid wsgi server for development purpose
    """
    load_init_function_from_entry_points()
    argv = [] + sys.argv
    Configuration.load('pyramid')
    sys.argv = argv
    configuration_post_load()
    BlokManager.load()
    config = Configurator()
    config.include_from_entry_point()
    config.load_config_bloks()

    wsgi_host = Configuration.get('wsgi_host')
    wsgi_port = int(Configuration.get('wsgi_port'))

    reload_at_change = Configuration.get('pyramid.reload_all', False)
    if reload_at_change:
        hupper.start_reloader('anyblok_pyramid.scripts.wsgi')

    app = config.make_wsgi_app()
    server = make_server(wsgi_host, wsgi_port, app)
    preload_databases(loadwithoutmigration=False)

    try:
        from colorama import Fore, Style
        start_msg = (
            "Pyramid development server running at %shttp://%s:%s%s" % (
                Fore.BLUE, wsgi_host, wsgi_port, Style.RESET_ALL))
    except ImportError:
        logger.info("`pip install colorama` to get colored link")
        start_msg = "Pyramid development server running at http://%s:%s" % (
            wsgi_host, wsgi_port)

    logger.info(start_msg)
    print(start_msg)

    server.serve_forever()


def gunicorn_wsgi():
    """console script function to run anyblok / pyramid with gunicorn"""
    try:
        import gunicorn  # noqa
    except ImportError:
        logger.error("Gunicorn is not installed. Try: pip install gunicorn")
        sys.exit(1)

    from .gunicorn import WSGIApplication
    WSGIApplication('gunicorn').run()



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Alexis TOURNEUX <tourneuxalexis@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytz
import time
from base64 import b64encode
from decimal import Decimal
from datetime import timedelta

from enum import Enum, unique


def datetime_adapter(obj, request):
    """Format the fields.DateTime to return String

    If the datetime hasn't any timezone, force the timezone by
    the server timezone

    ::

        from pyramid.renderers import JSON
        from datetime import datetime
        json_renderer = JSON()
        json_renderer.add_adapter(datetime, datetime_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: datetime obj
    :rtype: str, isoformat datetime
    """
    if obj is not None:
        if obj.tzinfo is None:
            timezone = pytz.timezone(time.tzname[0])
            obj = timezone.localize(obj)

    return obj.isoformat()


@unique
class TimeDeltaModes(Enum):

    """This enum is aimed at creating constants for setting the serialization
       mode for timedelta adapter."""

    MICROSECONDS = 1
    MILLISECONDS = 2
    SECONDS = 3
    MINUTES = 4
    HOURS = 5
    DAYS = 6
    WEEKS = 7


def timedelta_adapter_factory(mode=TimeDeltaModes.SECONDS):
    def timedelta_adapter(obj, request):

        """Format the fields.TimeDelta to return String

        ::
            from pyramid.renderers import JSON
            from datetime import timedelta
            from anyblok_pyramid.adapter import timedelta_adapter_factory
            json_renderer = JSON()
            json_renderer.add_adapter(timedelta, timedelta_adapter_factory())
            config.add_renderer('json', json_renderer)

        The mode parameter can be used to set the output of the adapter.
        This parameter can either be :

            * microseconds -> TimeDeltaModes.MICROSECONDS
            * milliseconds -> TimeDeltaModes.MILLISECONDS
            * seconds (default) -> TimeDeltaModes.SECONDS
            * minutes -> TimeDeltaModes.MINUTES
            * hours -> TimeDeltaModes.HOURS
            * days -> TimeDeltaModes.DAYS
            * weeks -> TimeDeltaModes.WEEKS

        When using this adapter with non default parameter, it might be used the
        following way :

        ::
            from pyramid.renderers import JSON
            from datetime import timedelta
            from anyblok_pyramid.adapter import (
                timedelta_adapter_factory,
                TimeDeltaModes
            )

            json_renderer = JSON()
            json_renderer.add_adapter(
                timedelta, timedelta_adapter_factory(TimeDeltaModes.HOURS))
            config.add_renderer('json', json_renderer)

        :param obj: timedelta obj
        :param mode: str
        :rtype: str, seconds corresponding to timedelta
        """

        if mode == TimeDeltaModes.MICROSECONDS:
            return obj / timedelta(microseconds=1)
        elif mode == TimeDeltaModes.MILLISECONDS:
            return obj / timedelta(milliseconds=1)
        elif mode == TimeDeltaModes.SECONDS:
            return obj.total_seconds()
        elif mode == TimeDeltaModes.MINUTES:
            return obj / timedelta(minutes=1)
        elif mode == TimeDeltaModes.HOURS:
            return obj / timedelta(hours=1)
        elif mode == TimeDeltaModes.DAYS:
            return obj / timedelta(days=1)
        elif mode == TimeDeltaModes.WEEKS:
            return obj / timedelta(weeks=1)
        else:
            raise ValueError(
                ("Provided mode for timedelta_adapter is not valid. Found '%s'"
                 "." % mode))

    return timedelta_adapter


def date_adapter(obj, request):
    """Format the fields.Date to return String

    ::

        from pyramid.renderers import JSON
        from datetime import date
        json_renderer = JSON()
        json_renderer.add_adapter(date, date_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: datetime obj
    :rtype: str, isoformat datetime
    """
    return obj.isoformat()


def uuid_adapter(obj, request):
    """Format the fields.UUID to return String

    ::

        from pyramid.renderers import JSON
        from uuid import UUID
        json_renderer = JSON()
        json_renderer.add_adapter(UUID, uuid_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: uuid obj
    :rtype: str
    """
    return str(obj.hex)


def bytes_adapter(obj, request):
    """Format the fields.Binary to return String

    ::

        from pyramid.renderers import JSON
        json_renderer = JSON()
        json_renderer.add_adapter(bytes, bytes_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: bytes
    :rtype: str
    """
    return b64encode(obj).decode("utf-8")


def decimal_adapter(obj, request):
    """Format the fields.Decimal to return String

    ::

        from pyramid.renderers import JSON
        json_renderer = JSON()
        json_renderer.add_adapter(Decimal, decimal_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: Decimal
    :rtype: str
    """
    return str(Decimal(obj).quantize(Decimal('1.00')))


def enum_adapter(obj, request):
    """Format the fields.Enum to return String

    ::

        from pyramid.renderers import JSON
        import enum

        json_renderer = JSON()
        json_renderer.add_adapter(enum.Enum, enum_adapter)
        config.add_renderer('json', json_renderer)


    :param obj: Enum
    :rtype: str
    """
    return obj.value



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok


class Test(Blok):

    version = '1.0.0'

    @classmethod
    def pyramid_load_config(cls, config):
        config.add_route('hello', '/hello/{name}/')
        config.scan(cls.__module__ + '.views')



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.config import (
    define_wsgi_option,
    define_auth_option,
    define_wsgi_debug_option,
    add_configuration_file,
    update_plugins
)
from anyblok.tests.test_config import MockArgumentParser


class TestArgsParseOption:

    def test_define_wsgi_option(self):
        parser = MockArgumentParser()
        define_wsgi_option(parser)

    def test_define_auth_option(self):
        parser = MockArgumentParser()
        define_auth_option(parser)

    def test_define_wsgi_debug_option(self):
        parser = MockArgumentParser()
        define_wsgi_debug_option(parser)

    def test_add_configuration_file(self):
        parser = MockArgumentParser()
        add_configuration_file(parser)

    def test_update_plugins(self):
        parser = MockArgumentParser()
        update_plugins(parser)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
try:
    from pyramid.authorization import ACLAuthorizationPolicy
except ImportError:
    from pyramid.security import ACLAuthorizationPolicy  # pragma: no cover


from anyblok_pyramid.security import RootFactory
from .pyramid import getAuthenticationPolicy


def declarations(reload=None):
    from . import model
    if reload:
        reload(model)


class Pyramid(Blok):
    version = "0.1.0"
    author = "Jean-Sébastien Suzanne"
    required = ['anyblok-core']

    @classmethod
    def import_declaration_module(cls):
        declarations()

    @classmethod
    def reload_declaration_module(cls, reload):
        declarations(reload=reload)

    @classmethod
    def pyramid_load_config(cls, config):
        # get authentication on view
        authn_policy = getAuthenticationPolicy()
        config.set_authentication_policy(authn_policy)
        # get autorization
        authz_policy = ACLAuthorizationPolicy()
        config.set_authorization_policy(authz_policy)
        config.set_root_factory(RootFactory)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import Integer, String, Json
from anyblok.relationship import Many2One
from anyblok.field import JsonRelated
from .exceptions import AuthorizationValidationException
from sqlalchemy import or_
try:
    from pyramid.authorization import Allow, Deny, ALL_PERMISSIONS
except ImportError:
    from pyramid.security import Allow, Deny, ALL_PERMISSIONS


Pyramid = Declarations.Model.Pyramid


@Declarations.register(Pyramid)
class Authorization:
    """A model to store autorization rules (permissions for users against an
    Anyblok model or a Pyramid resource)"""

    id = Integer(primary_key=True)
    code = String(nullable=True, unique=True, size=256)
    order = Integer(default=100, nullable=False)

    resource = String()
    model = String(
        foreign_key=Declarations.Model.System.Model.use('name').options(
            ondelete="cascade"),
        size=256,
    )
    primary_keys = Json(default={})
    filter = Json(default={})  # next step

    role = Many2One(
        model=Pyramid.Role, foreign_key_options={'ondelete': 'cascade'})
    login = String(
        foreign_key=Pyramid.User.use('login').options(ondelete="cascade"))
    user = Many2One(model=Pyramid.User)
    perms = Json(default={})

    perm_create = JsonRelated(json_column='perms', keys=['create'])
    perm_read = JsonRelated(json_column='perms', keys=['read'])
    perm_update = JsonRelated(json_column='perms', keys=['update'])
    perm_delete = JsonRelated(json_column='perms', keys=['delete'])

    @classmethod
    def get_acl_filter_model(cls):
        """Return the Model to use to check the permission"""
        return {
            'User': cls.anyblok.Pyramid.User,
            'Role': cls.anyblok.Pyramid.Role,
        }

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Return the Pyramid ACL in function of the resource and user

        :param login: str, login of the user
        :param resource: str, name of the resource
        """
        # cache the method
        User = cls.anyblok.Pyramid.User
        Role = cls.anyblok.Pyramid.Role

        query = cls.query()
        query = query.filter(
            or_(cls.resource == resource, cls.model == resource))
        query = query.order_by(cls.order)
        Q1 = query.filter(cls.login == login)
        Q2 = query.join(cls.role).filter(Role.name.in_(User.get_roles(login)))
        res = []
        for query in (Q1, Q2):
            for self in query.all():
                allow_perms = []
                deny_perms = []
                perms = list((self.perms or {}).keys())
                perms.sort()
                for perm in perms:
                    p = self.perms[perm]
                    query = User.query()
                    query = query.filter(User.login == login)
                    query = query.join(User.roles)
                    if self.filter:
                        query = query.condition_filter(
                            self.filter,
                            cls.get_acl_filter_model()
                        )

                    if 'condition' in p:
                        query = query.condition_filter(
                            p['condition'],
                            cls.get_acl_filter_model()
                        )

                    ismatched = True if query.count() else False
                    if p.get('matched' if ismatched else 'unmatched') is True:
                        allow_perms.append(perm)
                    elif (
                        p.get('matched' if ismatched else 'unmatched') is False
                    ):
                        deny_perms.append(perm)

                if len(allow_perms):
                    res.append((Allow, login, allow_perms))

                if len(deny_perms):
                    res.append((Deny, login, deny_perms))

        res.append((Deny, login, ALL_PERMISSIONS))
        return res

    @classmethod
    def check_acl(cls, login, resource, type_):
        """Return the Pyramid ACL in function of the resource and user

        :param login: str, login of the user
        :param resource: str, name of the resource
        :param type: str, name of the action
        """
        # cache the method
        User = cls.anyblok.Pyramid.User
        Role = cls.anyblok.Pyramid.Role

        query = cls.query()
        query = query.filter(
            or_(cls.resource == resource, cls.model == resource))
        query = query.order_by(cls.order)
        Q1 = query.filter(cls.login == login)
        Q2 = query.join(cls.role).filter(Role.name.in_(User.get_roles(login)))
        for query in (Q1, Q2):
            for self in query.all():
                perms = list((self.perms or {}).keys())
                if type_ not in perms:
                    continue

                p = self.perms[type_]
                query = User.query()
                query = query.filter(User.login == login)
                query = query.join(User.roles, isouter=True)
                if self.filter:
                    query = query.condition_filter(
                        self.filter,
                        cls.get_acl_filter_model()
                    )

                if 'condition' in p:
                    query = query.condition_filter(
                        p['condition'],
                        cls.get_acl_filter_model()
                    )

                ismatched = True if query.count() else False
                if p.get('matched' if ismatched else 'unmatched') is True:
                    return True
                elif (
                    p.get('matched' if ismatched else 'unmatched') is False
                ):
                    return False

        return False

    @classmethod
    def before_insert_orm_event(cls, mapper, connection, target):
        target.check_validity()

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        target.check_validity()

    def check_validity(self):
        """When creating or updating a User.Authorization, check that all rules
        objects exists or return an AuthorizationValidationException

        :exception: AuthorizationValidationException
        """
        if not (self.role or self.login or self.user or self.role_name):
            raise AuthorizationValidationException(
                "No role and login to apply in the authorization (%s)" % self)

        if not (self.resource or self.model):
            raise AuthorizationValidationException(
                "No resource and model to apply in the authorization (%s)" %
                self)

        if not self.model and self.primary_keys:
            raise AuthorizationValidationException(
                "Primary keys without model to apply in the authorization "
                "(%s)" % self)

    @classmethod
    def ensure_exists(
        cls, code, **kwargs
    ):
        """Ensure role's authorization is present

        :param code: String, authorization code.
        :param kwargs: authorization fields
        """
        if not kwargs:
            kwargs = {}  # pragma: no cover

        # pv: at some point adding index on this criteria may boost things
        # while setting authorizations
        authz = (
            cls.anyblok.Pyramid.Authorization.query()
            .filter_by(
                code=code,
            )
            .one_or_none()
        )
        if not authz:
            authz = cls.anyblok.Pyramid.Authorization.insert(
                code=code, **kwargs
            )
        else:
            authz.update(**kwargs)
            jsonfields = {"perms", "primary_keys", "filter"}
            perms_related = {
                "perm_create",
                "perm_read",
                "perm_update",
                "perm_delete"
            }

            modified = (jsonfields | perms_related) & set(kwargs.keys())
            if modified:
                if perms_related & modified:
                    modified = modified | {"perms"}
                authz.flag_modified(*(jsonfields & modified))
        return authz


@Declarations.register(Pyramid)
class Role:

    @classmethod
    def ensure_exists(
        cls, name, authorizations, label=None
    ):
        """Create or update Pyramid.Role with related model's authorization.

        :param name: str, Role name
        :param authorizations: list, `Model.Pyramid.Authorization` data to
            ensure that it exists::

            [
                {
                    # if code is not provide it will raise an exception
                    "code": "Uniq auth code by role",
                    "model": "Model.Test",
                    "perms": merge(
                        PERM_WRITE,
                        "whatever you needs": {"matched": True}
                    },
                    "order": 100,
                    # Any other values present on Model.Pyramid.Authorization
                },
                {
                    # if code is not provide it will raise an exception
                    "code": "Uniq auth code by role",
                    "resource": "a resource",
                    "perms": {
                        "hack_resource": {"matched": True},
                    },
                },
                ...
            ]

        :param label: str, Role label used if role doesn't exits on insert.
                      default: capitalized name.
        :return: Created or updated role
        """
        Pyramid = cls.anyblok.Pyramid
        role = Pyramid.Role.query().get(name)
        if not role:
            if not label:
                label = name.capitalize()
            role = Pyramid.Role.insert(name=name, label=label)

        for authz in authorizations:
            code = authz.pop("code", None)
            if code is None:
                raise KeyError(
                    f"Missing `code` information to ensure "
                    f"Model.Pyramid.Authorizatoin is present on role "
                    f"{role.name}"
                )
            Pyramid.Authorization.ensure_exists(
                code, role=role, **authz
            )
        return role



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pyramid.httpexceptions import HTTPUnauthorized
from anyblok import Declarations
from anyblok.column import String
try:
    from pyramid.authorization import Allow, ALL_PERMISSIONS
except ImportError:
    from pyramid.security import Allow, ALL_PERMISSIONS


@Declarations.register(Declarations.Model)
class Pyramid:

    @classmethod
    def get_roles(cls, login):
        """Return the roles of an user

        :param login: str, login attribute of the user
        :rtype: list of str (name of the roles)
        """
        return cls.anyblok.Pyramid.User.get_roles(login)

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Retun the ACL for a ressource and a user

        Auth, does not implement any rule to compute ACL,
        This method allow all user to use the resource ask
        by controllers.

        For other configuration, this method must be overwrite

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param params: all options need to compute ACL
        """
        return cls.anyblok.Pyramid.User.get_acl(
            login, resource, params=params)

    @classmethod
    def check_acl(cls, login, resource, type_):
        """Retun True if user is allowed to make action type
        of the resource

        This method must be ober writting by the auth blok

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param type: str, name of the action
        :param params: all options need to compute ACL
        """
        return cls.anyblok.Pyramid.User.check_acl(
            login, resource, type_)

    @classmethod
    def check_login(cls, **kwargs):
        """Check login / password

        This method raise an exception, because any credential
        is stored in this bloks

        .. warning::

            This method must be overwriting by anycredential blok


        :param kwargs: any options need to validate credential
        """
        return cls.anyblok.Pyramid.User.check_login(**kwargs)

    @classmethod
    def check_user_exists(cls, login):
        user = cls.anyblok.Pyramid.User.query().get(login)
        if user is None:
            raise KeyError(f"{login} is not a valid login")

        return user

    @classmethod
    def get_user(cls, user_id):
        """Return user for a given user_id.
        The method is called by `Model.Pyramid.get_user` cached method
        to retreive user. You (as developer) must implement a cache
        invalidation in case of user modification that could impact
        restricted query by user id
        """
        return cls.anyblok.Pyramid.User.query().get(user_id)


@Declarations.register(Declarations.Model.Pyramid)
class User:
    """User declaration need for Auth"""

    login = String(primary_key=True, nullable=False)

    @classmethod
    def get_roles(cls, login):
        """Return the roles of an user

        :param login: str, login attribute of the user
        :rtype: list of str (name of the roles)
        """
        # cache the method
        roles = [login]
        user = cls.query().filter(cls.login == login).one_or_none()
        if user:
            for role in user.roles:
                roles.extend(role.roles_name)

        return list(set(roles))

    @classmethod
    def get_acl(cls, login, resource, params=None):
        """Retun the ACL for a ressource and a user

        Auth, does not implement any rule to compute ACL,
        This method allow all user to use the resource ask
        by controllers.

        For other configuration, this method must be overwrite

        :param login: str, login attribute of the user
        :param resource: str, name of a resource
        :param params: all options need to compute ACL
        """
        # cache the method
        return [(Allow, login, ALL_PERMISSIONS)]

    @classmethod
    def check_acl(cls, login, resource, type_):
        """Overwrite the method to return the ACL for the resource and user

        :param login: str, login of the user
        :param resource: str, name of the resource
        :param type: str, name of the action
        """
        return True

    @classmethod
    def check_login(cls, login=None, password=None, **kwargs):
        """Check login / password

        This method raise an exception, because any credential
        is stored in this bloks

        .. warning::

            This method must be overwriting by anycredential blok


        :param login: str, the login attribute of the user
        :param password: str
        :param kwargs: any options need to validate credential
        """
        raise HTTPUnauthorized()

    @classmethod
    def get_login_location_to(cls, login, request):
        """Return the default path after the login"""
        return '/'

    @classmethod
    def get_logout_location_to(cls, request):
        """Return the default path after the logout"""
        return '/'



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2002 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok_pyramid.bloks.pyramid.restrict import restrict_query_by_user
try:
    from pyramid.authorization import Allow, Authenticated
except ImportError:
    from pyramid.security import Allow, Authenticated


@Declarations.register(Declarations.Model.Pyramid)
class User:

    @classmethod
    def check_login(cls, login=None, password=None):
        return cls.anyblok.Pyramid.User.query().get(login)

    @classmethod
    def get_acl(cls, login, resource, **params):
        return [
            (Allow, Authenticated, 'read'),
            (Allow, 'admin', 'write'),
        ]


@Declarations.register(Declarations.Model)
class Pyramid:
    @classmethod
    def check_user_exists(cls, login):
        # if login in ["user@anyblok.org", "user2@anyblok.org"]:
        if login == "user@anyblok.org":
            return
        raise ValueError("Unknown user")


@Declarations.register(Declarations.Model.System)
class Blok:

    @restrict_query_by_user()
    def restrict_reading_this_blok_to_user2(cls, query, user):
        if user.login == "user2@anyblok.org":
            query = query.filter_by(name="test-pyramid2")
        return query



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2002 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok_pyramid.security import AnyBlokResourceFactory
from anyblok_pyramid.bloks.auth.views import login, logout


try:
    from anyblok_pyramid.bloks.pyramid.oidc import (
        login as oidc_login,
        callback as oidc_callback,
    )
    has_oidc = True
except ImportError:
    has_oidc = False


class Test(Blok):

    version = "1.0.0"
    required = ["auth"]

    @classmethod
    def import_declaration_module(cls):
        from . import models  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import models
        reload(models)

    @classmethod
    def pyramid_load_config(cls, config):
        config.add_route(
            "bloks",
            "/bloks",
            factory=AnyBlokResourceFactory("Model.System.Blok"),
        )
        config.add_route(
            "blok",
            "/blok/{name}",
            factory=AnyBlokResourceFactory("Model.System.Blok"),
        )
        config.add_route("login", "/login", request_method="POST")
        config.add_view(view=login, route_name="login", renderer="JSON")
        config.add_route("logout", "/logout", request_method="POST")
        config.add_view(view=logout, route_name="logout")

        if has_oidc:
            config.add_route("oidc_login", "/oidc_login", request_method="GET")
            config.add_view(view=oidc_login, route_name="oidc_login")
            config.add_route(
                "oidc_callback", "/oidc_callback", request_method="GET"
            )
            config.add_view(view=oidc_callback, route_name="oidc_callback")

        config.scan(cls.__module__ + ".views")

    def update(self, latest):
        if not latest:
            self.anyblok.Pyramid.User.insert(login="admin")
            self.anyblok.Pyramid.User.insert(login="viewer")
            self.anyblok.Pyramid.User.insert(login="user@anyblok.org")
            self.anyblok.Pyramid.User.insert(login="user2@anyblok.org")



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import warnings
from anyblok.tests.testcase import DBTestCase, BlokTestCase
from webtest import TestApp
from .pyramid_config import Configurator


def init_web_server(*functions):
    config = Configurator()
    config.include_from_entry_point()
    for function in functions:
        config.include(function)

    config.load_config_bloks()
    app = config.make_wsgi_app()
    return TestApp(app)


class PyramidTestCase:

    @classmethod
    def setUpClass(cls):
        warnings.warn(
            "The testCase from anyblok_pyramid for nose are deprecated. "
            "Refactor all your unittest with pytest: %r" % cls,
            DeprecationWarning, stacklevel=2)
        super(PyramidTestCase, cls).setUpClass()

    def setUp(self):
        super(PyramidTestCase, self).setUp()
        self.includemes = []

    def init_web_server(self):
        return init_web_server(*self.includemes)


class PyramidDBTestCase(PyramidTestCase, DBTestCase):

    def init_registry_with_bloks(self, *args, **kwargs):
        res = super(PyramidDBTestCase, self).init_registry_with_bloks(
            *args, **kwargs)
        self.webserver = self.init_web_server()
        return res


class PyramidBlokTestCase(PyramidTestCase, BlokTestCase):

    @property
    def webserver(self):
        return self.init_web_server()



"""This module provide OIDC Relying Party utilities

The service you are developing is the `Relying Party` (RP) in OIDC
terms. This service will delagate the authentication to an OIDC Provider also
called Issuer.

Currently only **Authorization Code Flow** is supported.

According `specifications <https://openid.net/specs/
openid-connect-core-1_0.html#CodeFlowSteps>`_

The Authorization Code Flow goes through the following steps:

  * The RP prepares an Authentication Request containing the desired request
    parameters.
  * User is redirect to the the Authorization Server (OIDC Provider/ Issuer).
  * Authorization Server Authenticates the End-User.
  * Authorization Server obtains End-User Consent/Authorization.
  * Authorization Server sends the End-User back to the RP with an
    Authorization Code.
  * RP requests a response using the Authorization Code at the Token Endpoint
    (OIDC Provider/ Issuer).
  * RP receives a response that contains an ID Token and Access Token in the
    response body.
  * RP validates the ID token and retrieves the End-User's Subject Identifier.

"""
try:
    import oic  # noqa
except ImportError as e:  # pragma: no cover
    raise ImportError('%s => pip install oic' % str(e))

from functools import lru_cache
from anyblok.config import Configuration
from oic import rndstr
from oic.oic import Client
from oic.oic.message import (
    AccessTokenResponse,
    AuthorizationResponse,
    RegistrationResponse,
)

from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from pyramid.httpexceptions import HTTPFound, HTTPUnauthorized
from pyramid.security import remember


@lru_cache()
def get_client():
    """Method that prepare the oidc client.

    This method use lru_cache in order call the OIDC provider once
    per thread

    Today we only support OIDC providers (ISSUER) that expose a
    /.well-known/openid-configuration route

    At the moment to authenticate the RP to the OIDC provider we only
    support through client_id/secret_ID/rp_callback attributes.

    You must configure OIDC in AnyBlok configuration:

    * **oidc-provider-issuer**: The OIDC Provider urls
        (ie: https://gitlab.com)
    * **oidc-relying-party-callback**: The Relaying Party callback, once
        the user is authenticate against the OIDC provider he will be redirect
        to that uri to the RP service (ie: http://localhost:8080/callback).
        In general this value is also configured in your OIDC provider to
        avoid redirection issues.
    * **oidc-relying-party-client-id**: The client id provide by your OIDC
        provider
    * **oidc-relying-party-secret-id**: The secret id provide by your OIDC
        provider

    And optionally:

    * **oidc-scope**: Specify what access privileges are being requested for
        Access Tokens. `cf Requesting claims using scope values
        <https://openid.net/specs/openid-connect-core-1_0.html#ScopeClaims`_.
        a list of claims using coma separator. (default value: `openid,email`)
    * **oidc-userinfo-field**: Specify which field to use from the response
        of the OIDC provider `userinfo endpoint <https://openid.net/specs/
        openid-connect-core-1_0.html#UserInfoResponse>`_. To make sure
        it's a known user. (default value: `email`).
    """

    client = Client(client_authn_method=CLIENT_AUTHN_METHOD)
    for config in [
        "oidc_provider_issuer",
        "oidc_relying_party_client_id",
        "oidc_relying_party_secret_id",
        "oidc_relying_party_callback",
    ]:
        if Configuration.get(config, None) is None:
            raise ValueError("You must provide {} parameter".format(config))
    provider_info = client.provider_config(
        Configuration.get("oidc_provider_issuer")
    )
    info = {
        "client_id": Configuration.get("oidc_relying_party_client_id"),
        "client_secret": Configuration.get("oidc_relying_party_secret_id"),
        "redirect_uris": [Configuration.get("oidc_relying_party_callback")],
    }
    info.update(provider_info._dict)
    client_reg = RegistrationResponse(**info)

    client.store_registration_info(client_reg)
    return client


def prepare_auth_url(request):
    """Prepare redirect uri to the OIDC provider, return headers to add
    a cookie regarding the current anonymous session in order to retreive
    the according generated state

    You may use it likes this (using cornice)::


        from cornice import Service
        from anyblok_pyramid.bloks.pyramid import oidc
        from pyramid.httpexceptions import HTTPFound

        login_route = Service(
            name='login',
            path='/login',
            installed_blok=current_blok()
        )

        @login_route.get()
        def login(request):
            HTTPFound(
                location=oidc.prepare_auth_url(request), headers=headers
            )
    """
    # Makes sure request.session is there what ever the implemented session
    # This should implements a Pyramid Idict interface
    if not getattr(request, "session", None):
        raise RuntimeError(  # pragma: no cover
            "In order to use OIDC Relaying party utility, "
            "you must configure a Pyramid session "
            "object https://docs.pylonsproject.org/projects/pyramid/"
            "en/latest/narr/sessions.html"
        )
    client = get_client()
    # headers = remember(request, "anonymous")
    request.session.update({"oic_state": rndstr()})

    args = {
        "client_id": client.client_id,
        "response_type": "code",
        "scope": Configuration.get("oidc_scope", "").split(","),
        # nonce is a string value used to associate a Client session
        # with an ID Token, and to mitigate replay attacks.
        "nonce": rndstr(),
        "redirect_uri": client.registration_response["redirect_uris"][0],
        "state": request.session["oic_state"],
    }
    auth_req = client.construct_AuthorizationRequest(request_args=args)
    login_url = auth_req.request(client.authorization_endpoint)
    return login_url  # , headers


def validate_response(response):
    """Make sure response is an instance of an AuthorizationResponse or
    an ErrorResponse.
    """
    if not isinstance(response, AuthorizationResponse):
        # it should be an ErrorResponse
        raise ValueError(  # pragma: no cover
            "Something goes wrong on the OIDC Provider: {}".format(response)
        )

    return response


def validate_state(request, response):
    """State is generated on the first call before redirection to the OIDC
    provider and must be the same, when user comme back from the OIDC
    provider.

    It is used to keep track of responses to outstanding requests (state).
    """
    if not response.get("state") == request.session.get("oic_state"):
        raise ValueError(  # pragma: no cover
            "State must be the same between registration and callback"
        )


def get_access_token(response):
    """Request an access token to retreive user info on OIDC server side
    """
    args = {"code": response["code"]}
    atr = get_client().do_access_token_request(
        state=response["state"],
        request_args=args,
        authn_method="client_secret_basic",
    )
    if not isinstance(atr, AccessTokenResponse):
        # it should be a TokenErrorResponse
        raise ValueError(  # pragma: no cover
            "OIDC Access token is invalid: {}".format(atr))
    return atr


def get_token(request):
    """Get a token in order to retreive data from the OIDC provider"""
    response = get_client().parse_response(
        AuthorizationResponse, info=request.query_string, sformat="urlencoded",
    )
    validate_response(response)
    validate_state(request, response)
    access_token_response = get_access_token(response)
    return response, access_token_response


def get_user_info(response):
    """Request to the OIDC provider to get user informations according the
    requested scope"""
    return get_client().do_user_info_request(state=response["state"])


def log_user(request):
    """Validate OIDC code and request info to validate if user exists

    invalidate current session if Pyramid.check_user_exists doesn't raise
    and return (userinfo, headers)
    """
    response, _ = get_token(request)
    userinfo = get_user_info(response)
    login = userinfo[Configuration.get("oidc_userinfo_field", None)]
    try:
        request.anyblok.registry.Pyramid.check_user_exists(login)
    except Exception:
        # TODO provide a better default way to manage invalid users
        request.anyblok.registry.rollback()
        return userinfo, None

    # Renew the Session ID After Privilege Level Change
    # This is fore security reason cf `OWASP cheat sheet
    # <https://cheatsheetseries.owasp.org/
    # cheatsheets/Session_Management_Cheat_Sheet.html
    # #renew-the-session-id-after-any-privilege-level-change`_
    request.session.invalidate()
    return userinfo, remember(request, login)


def login(request):
    """Default OIDC login route to prepare redirection to OIDC Provider to
    let user authenticate against. """
    return HTTPFound(location=prepare_auth_url(request))


def callback(request):
    """Default callback login route

    Once OIDC info are retrive, check if users exists according data
    return by the OIDC provider and renew session ID, then redirect
    user to the  Pyramid.User.get_login_location_to definition and set the
    session ID cookie"""
    userinfo, headers = log_user(request)
    if headers is None:
        raise HTTPUnauthorized()
    login = userinfo[Configuration.get("oidc_userinfo_field", None)]
    User = request.anyblok.registry.Pyramid.User
    location = User.get_login_location_to(login, request)
    return HTTPFound(location=location, headers=headers)



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from sqlalchemy.sql.elements import BooleanClauseList

from anyblok.column import String
from anyblok.declarations import Declarations
from anyblok.tests.conftest import init_registry_with_bloks

from anyblok_pyramid.bloks.pyramid.restrict import restrict_query_by_user

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


@pytest.fixture
def setup_registry(request, bloks_loaded):
    def setup(bloks, function, **kwargs):
        registry = init_registry_with_bloks(bloks, function, **kwargs)
        request.addfinalizer(registry.close)
        return registry

    return setup


def add_in_registry_inherited(with_super=False):
    @register(Mixin)
    class MTest:
        @restrict_query_by_user()
        def my_restrict_query_by_user_method(cls, query, user):
            return query.filter_by(name="overriden")

        @restrict_query_by_user()
        def my_mixin_restrict_query_by_user_method(cls, query, user):
            return query.filter_by(name="mixin")

    @register(Model)
    class RestrictedModel(Mixin.MTest):

        name = String(primary_key=True, nullable=False)

        @restrict_query_by_user()
        def my_restrict_query_by_user_method(cls, query, user):
            return query.filter_by(name="base")

    @register(Model)  # noqa: F811
    class RestrictedModel:  # noqa: F811
        @restrict_query_by_user()
        def my_restrict_query_by_user_method(cls, query, user):
            if with_super:
                query = super().my_restrict_query_by_user_method(query, user)
            return query.filter_by(name="child")

        @restrict_query_by_user()
        def my_other_restrict_query_by_user_method(cls, query, user):
            return query.filter_by(name="other")


def get_filters_values(query):
    """Method that return a list of right clauses filters from a SQLAlechemy
    query"""
    right_clauses = []
    for c in query.clauses:
        if isinstance(c, BooleanClauseList):
            right_clauses.extend(get_filters_values(c))
        else:
            right_clauses.append(c.right.value)
    return sorted(right_clauses)


def test_restrict_query_by_user_inheritance(setup_registry):
    registry = setup_registry(
        ["pyramid", "anyblok-test"],
        add_in_registry_inherited,
        with_super=False,
    )
    query = registry.Pyramid.restrict_query_by_user(
        registry.RestrictedModel.query(), "test",
    )
    assert get_filters_values(query.whereclause) == ["child", "mixin", "other"]


def test_restrict_query_by_user_inheritance_calling_parent(setup_registry):
    registry = setup_registry(
        ["pyramid", "anyblok-test"], add_in_registry_inherited, with_super=True
    )
    query = registry.Pyramid.restrict_query_by_user(
        registry.RestrictedModel.query(), "test",
    )
    assert get_filters_values(query.whereclause) == [
        "base", "child", "mixin", "other",
    ]



# This file is a part of the AnyBlok / Pyramid project
#
#    Copyright (C) 2020 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok_pyramid import merge, PERM_READ, PERM_WRITE
from sqlalchemy.orm.exc import MultipleResultsFound


@pytest.mark.usefixtures("rollback_registry")
class TestAuthorization:
    @pytest.fixture(scope="function", autouse=True)
    def setUp(self, rollback_registry):
        self.registry = rollback_registry
        self.Authorization = self.registry.Pyramid.Authorization

    def assert_authz(self):
        test = self.registry.Pyramid.Role.query().get("test")
        assert test is not None
        assert test.label == "Test"
        assert self.Authorization.query().filter_by(role=test).count() == 2
        authz_blok = (
            self.Authorization.query()
            .filter_by(role=test, model="Model.System.Blok")
            .one()
        )
        assert authz_blok.order == 4
        assert authz_blok.perm_create is None
        assert authz_blok.perm_read == {"matched": True}
        assert authz_blok.perm_update is None
        assert authz_blok.perm_delete is None
        assert authz_blok.perms["custom_perm"] == {"matched": True}
        authz_user = (
            self.Authorization.query()
            .filter_by(role=test, model="Model.Pyramid.User")
            .one()
        )
        assert authz_user.perms == {
            "create": {"matched": True},
            "read": {"matched": True},
            "update": {"matched": True},
            "delete": {"matched": True},
        }

    def test_ensure_role_without_code(self):
        self.registry.Pyramid.User.insert(login="test")
        self.Authorization.ensure_exists(
            None, resource="test", login="test"
        )
        self.Authorization.insert(resource="test 2", login="test")
        with pytest.raises(MultipleResultsFound):
            self.Authorization.ensure_exists(
                None, resource="test", login="test"
            )

    def test_no_code_provide(self):
        with pytest.raises(KeyError) as err:
            self.registry.Pyramid.Role.ensure_exists(
                "test",
                [
                    {
                        "code": "test-blok-read",
                        "model": "Model.System.Blok",
                        "perms": merge(
                            PERM_READ, {"custom_perm": {"matched": True}}
                        ),
                        "order": 4,
                    },
                    {
                        "model": "Model.Pyramid.User",
                        "perms": PERM_WRITE,
                    },
                ],
                label="Test",
            )
        assert (
            "'Missing `code` information to ensure "
            "Model.Pyramid.Authorizatoin is present on role "
            "test'"
        ) == str(err.value)

    def test_ensure_role_insert(self):
        assert self.registry.Pyramid.Role.query().get("test") is None
        self.registry.Pyramid.Role.ensure_exists(
            "test",
            [
                {
                    "code": "test-blok-read",
                    "model": "Model.System.Blok",
                    "perms": merge(
                        PERM_READ, {"custom_perm": {"matched": True}}
                    ),
                    "order": 4,
                },
                {
                    "code": "test-user-write",
                    "model": "Model.Pyramid.User",
                    "perms": PERM_WRITE,
                },
            ],
            label="Test",
        )
        self.assert_authz()

    def test_update(self):
        count_before = self.Authorization.query().count()
        self.registry.Pyramid.Role.ensure_exists(
            "test",
            [
                {
                    "code": "test-blok-read",
                    "model": "Model.System.Blok",
                    "perms": merge(
                        PERM_WRITE, {"custom_perm": {"matched": True}}
                    ),
                    "order": 4,
                },
                {
                    "code": "test-user-write",
                    "model": "Model.Pyramid.User",
                    "perms": PERM_WRITE,
                },
            ],
        )
        assert self.Authorization.query().count() == count_before + 2
        self.registry.Pyramid.Role.ensure_exists(
            "test",
            [
                {
                    "code": "test-blok-read",
                    "model": "Model.System.Blok",
                    "perms": merge(
                        PERM_READ, {"custom_perm": {"matched": True}}
                    ),
                    "order": 4,
                },
            ],
        )
        assert self.Authorization.query().count() == count_before + 2
        self.assert_authz()

    def test_update2(self):
        self.registry.Pyramid.Role.ensure_exists(
            "test",
            [
                {
                    "code": "test-blok-read",
                    "model": "Model.System.Blok",
                    "perms": merge(
                        PERM_READ, {"custom_perm": {"matched": True}}
                    ),
                    "order": 4,
                },
                {
                    "code": "test-user-write",
                    "model": "Model.Pyramid.User",
                    "perms": PERM_READ,
                },
            ],
        )
        self.registry.Pyramid.Role.ensure_exists(
            "test",
            [
                {
                    "code": "test-user-write",
                    "model": "Model.Pyramid.User",
                    "perm_create": {"matched": True},
                    "perm_update": {"matched": True},
                    "perm_delete": {"matched": True},
                }
            ],
        )
        self.assert_authz()
